/*      */ package ucar.nc2.iosp.bufr.tables;
/*      */ 
/*      */ import java.io.BufferedReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.net.URL;
import java.nio.charset.Charset;
import java.util.ArrayList;
import java.util.Formatter;
import java.util.List;
import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.jdom.Document;
import org.jdom.Element;
import org.jdom.JDOMException;
import org.jdom.input.SAXBuilder;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import ucar.nc2.iosp.bufr.BufrIdentificationSection;
import ucar.nc2.iosp.bufr.Descriptor;
import ucar.nc2.util.TableParser;
import ucar.unidata.util.StringUtil;
/*      */ 
/*      */ 
/*      */ 
/*      */ 
/*      */ 
/*      */ 
/*      */ 
/*      */ 
/*      */ 
/*      */ 
/*      */ 
/*      */ 
/*      */ 
/*      */ 
/*      */ 
/*      */ 
/*      */ 
/*      */ 
/*      */ 
/*      */ 
/*      */ 
/*      */ 
/*      */ 
/*      */ 
/*      */ 
/*      */ 
/*      */ 
/*      */ 
/*      */ public class BufrTables
/*      */ {
/*      */   static final String RESOURCE_PATH = "/resources/bufrTables/";
/*  127 */   private static Logger log = LoggerFactory.getLogger(BufrTables.class);
/*      */   private static final boolean debugTable = false;
/*      */   private static final boolean showReadErrs = false;
/*      */   private static List<TableConfig> tables;
/*  133 */   private static Map<String, TableB> tablesB = new ConcurrentHashMap();
/*  134 */   private static Map<String, TableD> tablesD = new ConcurrentHashMap();
/*      */   private static final String canonicalLookup = "resource:/resources/bufrTables/local/tablelookup.csv";
/*  137 */   private static List<String> lookups = null;
/*      */ 
/*  398 */   private static String version13 = "wmo.v13.composite";
/*  399 */   private static String version14 = "wmo.v14";
/*      */ 
/*  990 */   private static final Pattern threeInts = Pattern.compile("^\\s*(\\d+)\\s+(\\d+)\\s+(\\d+)");
/*  991 */   private static final Pattern negOne = Pattern.compile("^\\s*-1");
/*      */ 
/*      */   public static void addLookupFile(String filename)
/*      */     throws FileNotFoundException
/*      */   {
/*  139 */     if (lookups == null) lookups = new ArrayList();
/*  140 */     File f = new File(filename);
/*  141 */     if (!f.exists()) throw new FileNotFoundException(new StringBuilder().append(filename).append(" not found").toString());
/*  142 */     lookups.add(filename);
/*      */   }
/*      */ 
/*      */   private static void readTableLookup() {
/*  146 */     tables = new ArrayList();
/*  147 */     if (lookups != null) {
/*  148 */       lookups.add("resource:/resources/bufrTables/local/tablelookup.csv");
/*  149 */       for (String fname : lookups)
/*  150 */         readTableLookup(fname);
/*      */     } else {
/*  152 */       readTableLookup("resource:/resources/bufrTables/local/tablelookup.csv");
/*      */     }
/*      */   }
/*      */ 
/*      */   private static void readTableLookup(String filename)
/*      */   {
/*      */     try
/*      */     {
/*  160 */       InputStream ios = openStream(filename);
/*  161 */       BufferedReader dataIS = new BufferedReader(new InputStreamReader(ios, Charset.forName("UTF8")));
/*  162 */       int count = 0;
/*      */       while (true) {
/*  164 */         String line = dataIS.readLine();
/*  165 */         if (line == null) break;
/*  166 */         if (!line.startsWith("#")) {
/*  167 */           count++;
/*      */ 
/*  169 */           String[] flds = line.split(",");
/*  170 */           if (flds.length >= 8)
/*      */           {
/*  175 */             int fldidx = 0;
/*      */             try {
                            /* 177 */TableConfig table = new TableConfig();
/*  178 */               table.center = Integer.parseInt(flds[(fldidx++)].trim());
/*  179 */               table.subcenter = Integer.parseInt(flds[(fldidx++)].trim());
/*  180 */               table.master = Integer.parseInt(flds[(fldidx++)].trim());
/*  181 */               table.local = Integer.parseInt(flds[(fldidx++)].trim());
/*  182 */               table.cat = Integer.parseInt(flds[(fldidx++)].trim());
/*  183 */               table.tableBname = flds[(fldidx++)].trim();
/*  184 */               table.tableBformat = flds[(fldidx++)].trim();
/*  185 */               table.tableDname = flds[(fldidx++)].trim();
/*  186 */               table.tableDformat = flds[(fldidx++)].trim();
/*  187 */               if (fldidx < flds.length) {
/*  188 */                 String modes = flds[(fldidx++)].trim();
/*  189 */                 if (modes.equalsIgnoreCase("wmoLocal"))
/*  190 */                   table.mode = Mode.wmoLocal;
/*  191 */                 else if (modes.equalsIgnoreCase("localWmo")) {
/*  192 */                   table.mode = Mode.localOverride;
/*      */                 }
/*      */               }
/*  195 */               tables.add(table);
/*      */             } catch (Exception e) {
/*      */             }
/*      */           }
/*      */         }
/*      */       }
/*  201 */       dataIS.close();
/*      */     } catch (IOException ioe) {
/*  203 */       String mess = new StringBuilder().append("Need BUFR tables in path; looking for ").append(filename).toString();
/*  204 */       throw new RuntimeException(mess, ioe);
/*      */     }
/*      */   }
/*      */ 
/*      */   private static TableConfig matchTableConfig(int center, int subcenter, int master, int local, int cat)
/*      */   {
/*  226 */     if (tables == null) readTableLookup();
/*      */ 
/*  228 */     for (TableConfig tc : tables) {
/*  229 */       if (tc.matches(center, subcenter, master, local, cat))
/*  230 */         return tc;
/*      */     }
/*  232 */     return null;
/*      */   }
/*      */ 
/*      */   private static TableConfig matchTableConfig(BufrIdentificationSection ids) {
/*  236 */     if (tables == null) readTableLookup();
/*      */ 
/*  238 */     int center = ids.getCenterId();
/*  239 */     int subcenter = ids.getSubCenterId();
/*  240 */     int master = ids.getMasterTableVersion();
/*  241 */     int local = ids.getLocalTableVersion();
/*  242 */     int cat = ids.getCategory();
/*      */ 
/*  244 */     return matchTableConfig(center, subcenter, master, local, cat);
/*      */   }
/*      */ 
/*      */   public static Tables getLocalTables(BufrIdentificationSection ids)
/*      */     throws IOException
/*      */   {
/*  368 */     TableConfig tc = matchTableConfig(ids);
/*  369 */     if (tc == null) return null;
/*      */ 
/*  371 */     if (tc.tableBformat.equals("ncep-nm"))
/*      */     {
/*  373 */       TableB b = (TableB)tablesB.get(tc.tableBname);
/*  374 */       TableD d = (TableD)tablesD.get(tc.tableBname);
/*  375 */       if ((b != null) && (d != null)) return new Tables(b, d, tc.mode);
/*      */ 
/*  378 */       b = new TableB(tc.tableBname, tc.tableBname);
/*  379 */       d = new TableD(tc.tableBname, tc.tableBname);
/*  380 */       Tables t = new Tables(b, d, tc.mode);
/*  381 */       InputStream ios = openStream(tc.tableBname);
/*  382 */       NcepMnemonic.read(ios, t);
/*      */ 
/*  385 */       tablesB.put(tc.tableBname, t.b);
/*  386 */       tablesD.put(tc.tableBname, t.d);
/*  387 */       return t;
/*      */     }
/*      */ 
/*  390 */     Tables tables = new Tables();
/*  391 */     tables.b = readTableB(tc.tableBname, tc.tableBformat, false);
/*  392 */     tables.d = readTableD(tc.tableDname, tc.tableDformat, false);
/*  393 */     tables.mode = tc.mode;
/*      */ 
/*  395 */     return tables;
/*      */   }
/*      */ 
/*      */   public static TableB getWmoTableB(BufrIdentificationSection ids)
/*      */     throws IOException
/*      */   {
/*  402 */     return getWmoTableB(ids.getMasterTableVersion());
/*      */   }
/*      */ 
/*      */   public static TableB getWmoTableB(int version) throws IOException {
        // String tableName = version == 14 ? version14 : version13;
        /*
         * Issue #2905 fixed table name to match version comparison logic used
         * below
         */
        /* 406 */String tableName = version < 14 ? version13 : version14;
/*  407 */     TableB tb = (TableB)tablesB.get(tableName);
/*  408 */     if (tb != null) return tb;
/*      */ 
/*  411 */     TableConfig tc14 = matchTableConfig(0, 0, 14, 0, -1);
/*  412 */     TableB result = readTableB(tc14.tableBname, tc14.tableBformat, false);
/*  413 */     tablesB.put(version14, result);
/*      */ 
/*  416 */     if (version < 14) {
/*  417 */       TableConfig tc = matchTableConfig(0, 0, 13, 0, -1);
/*  418 */       TableB b13 = readTableB(tc.tableBname, tc.tableBformat, false);
/*  419 */       TableB.Composite bb = new TableB.Composite(version13, version13);
/*  420 */       bb.addTable(b13);
/*  421 */       bb.addTable(result);
/*  422 */       result = bb;
/*  423 */       tablesB.put(version13, result);
/*      */     }
/*      */ 
/*  426 */     return result;
/*      */   }
/*      */ 
/*      */   public static TableB readTableB(String location, String format, boolean force) throws IOException {
/*  430 */     if (!force) {
/*  431 */       TableB tb = (TableB)tablesB.get(location);
/*  432 */       if (tb != null) return tb;
/*      */ 
/*      */     }
/*      */ 
/*  436 */     InputStream ios = openStream(location);
/*  437 */     TableB b = new TableB(location, location);
/*  438 */     if (format.equals("csv")) {
/*  439 */       readWmoTableB(ios, b);
/*  440 */     } else if (format.equals("ncep")) {
/*  441 */       readNcepTableB(ios, b);
/*  442 */     } else if (format.equals("ncep-nm")) {
/*  443 */       Tables t = new Tables(b, null, null);
/*  444 */       NcepMnemonic.read(ios, t);
/*  445 */     } else if (format.equals("ecmwf")) {
/*  446 */       readEcmwfTableB(ios, b);
/*  447 */     } else if (format.equals("ukmet")) {
/*  448 */       readBmetTableB(ios, b);
/*  449 */     } else if (format.equals("mel-bufr")) {
/*  450 */       readMelbufrTableB(ios, b);
/*  451 */     } else if (format.equals("mel-tabs")) {
/*  452 */       readMeltabTableB(ios, b);
/*  453 */     } else if (format.equals("wmo-xml")) {
/*  454 */       readWmoXmlTableB(ios, b);
/*      */     } else {
/*  456 */       System.out.printf("Unknown format= %s %n", new Object[] { format });
/*  457 */       return null;
/*      */     }
/*      */ 
/*  460 */     tablesB.put(location, b);
/*  461 */     return b;
/*      */   }
/*      */ 
/*      */   private static void readWmoTableB(InputStream ios, TableB b) throws IOException {
/*  465 */     BufferedReader dataIS = new BufferedReader(new InputStreamReader(ios, Charset.forName("UTF8")));
/*  466 */     int count = 0;
/*      */     while (true) {
/*  468 */       String line = dataIS.readLine();
/*  469 */       if (line == null) break;
/*  470 */       if (!line.startsWith("#")) {
/*  471 */         count++;
/*      */ 
/*  473 */         if (count != 1)
/*      */         {
/*  479 */           int pos1 = line.indexOf(34);
/*  480 */           if (pos1 >= 0) {
/*  481 */             int pos2 = line.indexOf(34, pos1 + 1);
/*  482 */             StringBuffer sb = new StringBuffer(line);
/*  483 */             for (int i = pos1; i < pos2; i++)
/*  484 */               if (sb.charAt(i) == ',') sb.setCharAt(i, ' ');
/*  485 */             line = sb.toString();
/*      */           }
/*      */ 
/*  488 */           String[] flds = line.split(",");
/*  489 */           if (flds.length >= 7)
/*      */           {
/*  494 */             int fldidx = 0;
/*      */             try {
/*  496 */               int classId = Integer.parseInt(flds[(fldidx++)].trim());
/*  497 */               int xy = Integer.parseInt(flds[(fldidx++)].trim());
/*  498 */               String name = StringUtil.remove(flds[(fldidx++)], 34);
/*  499 */               String units = StringUtil.filter(flds[(fldidx++)], " %+-_/()*");
/*  500 */               int scale = Integer.parseInt(clean(flds[(fldidx++)].trim()));
/*  501 */               int refVal = Integer.parseInt(clean(flds[(fldidx++)].trim()));
/*  502 */               int width = Integer.parseInt(clean(flds[(fldidx++)].trim()));
/*      */ 
/*  504 */               int x = xy / 1000;
/*  505 */               int y = xy % 1000;
/*      */ 
/*  507 */               b.addDescriptor((short)x, (short)y, scale, refVal, width, name, units); } catch (Exception e) {
/*      */             }
/*      */           }
/*      */         }
/*      */       }
/*      */     }
/*  513 */     dataIS.close();
/*      */   }
/*      */ 
/*      */   private static String clean(String s) {
/*  517 */     return StringUtil.remove(s, 32);
/*      */   }
/*      */ 
/*      */   private static TableB readMelbufrTableB(InputStream ios, TableB b)
/*      */     throws IOException
/*      */   {
/*  523 */     BufferedReader dataIS = new BufferedReader(new InputStreamReader(ios));
/*      */     while (true)
/*      */     {
/*  527 */       String line = dataIS.readLine();
/*  528 */       if (line == null) break;
/*  529 */       if ((!line.startsWith("#")) && (line.length() != 0))
/*      */       {
/*      */         try
/*      */         {
/*  534 */           String[] split = line.split(";");
/*  535 */           short x = Short.parseShort(split[1].trim());
/*  536 */           short y = Short.parseShort(split[2].trim());
/*  537 */           int scale = Integer.parseInt(split[3].trim());
/*  538 */           int refVal = Integer.parseInt(split[4].trim());
/*  539 */           int width = Integer.parseInt(split[5].trim());
/*      */ 
/*  541 */           b.addDescriptor(x, y, scale, refVal, width, split[7], split[6]);
/*      */         } catch (Exception e) {
/*  543 */           log.error(new StringBuilder().append("Bad table B entry: table=").append(b.getName()).append(" entry=<").append(line).append(">").toString(), e.getMessage());
/*      */         }
/*      */       }
/*      */     }
/*  547 */     dataIS.close();
/*      */ 
/*  549 */     return b;
/*      */   }
/*      */ 
/*      */   private static TableB readMeltabTableB(InputStream ios, TableB b)
/*      */     throws IOException
/*      */   {
/*  557 */     BufferedReader dataIS = new BufferedReader(new InputStreamReader(ios));
/*      */     while (true)
/*      */     {
/*  561 */       String line = dataIS.readLine();
/*  562 */       if (line == null) break;
/*  563 */       if ((!line.startsWith("#")) && (line.length() != 0))
/*      */       {
/*      */         try
/*      */         {
/*  567 */           String[] split = line.split("\t");
/*  568 */           short x = Short.parseShort(split[1].trim());
/*  569 */           short y = Short.parseShort(split[2].trim());
/*  570 */           int scale = Integer.parseInt(split[3].trim());
/*  571 */           int refVal = Integer.parseInt(split[4].trim());
/*  572 */           int width = Integer.parseInt(split[5].trim());
/*      */ 
/*  575 */           b.addDescriptor(x, y, scale, refVal, width, split[7], split[6]);
/*      */         } catch (Exception e) {
/*  577 */           log.error(new StringBuilder().append("Bad table ").append(b.getName()).append(" entry=<").append(line).append(">").toString(), e);
/*      */         }
/*      */       }
/*      */     }
/*  581 */     dataIS.close();
/*      */ 
/*  583 */     return b;
/*      */   }
/*      */ 
/*      */   private static TableB readNcepTableB(InputStream ios, TableB b)
/*      */     throws IOException
/*      */   {
/*  589 */     BufferedReader dataIS = new BufferedReader(new InputStreamReader(ios));
/*      */ 
/*  591 */     dataIS.readLine();
/*      */ 
/*  593 */     int count = 0;
/*      */     while (true) {
/*  595 */       String line = dataIS.readLine();
/*  596 */       if (line == null) break;
/*  597 */       if ((!line.startsWith("#")) && (line.length() != 0))
/*      */       {
/*      */         try
/*      */         {
/*  601 */           String[] flds = line.split("[\\|;]");
/*  602 */           if (flds[0].equals("END"))
/*      */             break;
/*  604 */           if (flds.length < 8) {
/*  605 */             log.error(new StringBuilder().append("Bad line in table ").append(b.getName()).append(" entry=<").append(line).append(">").toString());
/*      */           }
/*      */           else
/*      */           {
/*  609 */             String fxys = flds[0];
/*  610 */             int scale = Integer.parseInt(clean(flds[1]));
/*  611 */             int refVal = Integer.parseInt(clean(flds[2]));
/*  612 */             int width = Integer.parseInt(clean(flds[3]));
/*  613 */             String units = StringUtil.remove(flds[4], 34);
/*  614 */             String name = StringUtil.remove(flds[7], 34);
/*      */ 
/*  616 */             String[] xyflds = fxys.split("-");
/*  617 */             short x = Short.parseShort(clean(xyflds[1]));
/*  618 */             short y = Short.parseShort(clean(xyflds[2]));
/*      */ 
/*  620 */             b.addDescriptor(x, y, scale, refVal, width, name, units);
/*      */           }
/*      */ 
/*      */         }
/*      */         catch (Exception e)
/*      */         {
/*  628 */           log.error(new StringBuilder().append("Bad table ").append(b.getName()).append(" entry=<").append(line).append(">").toString(), e);
/*      */         }
/*      */       }
/*      */     }
/*      */ 
/*  633 */     dataIS.close();
/*      */ 
/*  635 */     return b;
/*      */   }
/*      */ 
/*      */   private static TableB readEcmwfTableB(InputStream ios, TableB b)
/*      */     throws IOException
/*      */   {
/*  645 */     int count = 0;
        /* 646 */List<TableParser.Record> recs = TableParser.readTable(ios,
                "4i,7i,72,97,102i,114i,119i", 50000);
/*  647 */     for (TableParser.Record record : recs) {
/*  648 */       if (record.nfields() >= 7)
/*      */       {
/*  651 */         int x = ((Integer)record.get(0)).intValue();
/*  652 */         int y = ((Integer)record.get(1)).intValue();
/*  653 */         String name = (String)record.get(2);
/*  654 */         String units = (String)record.get(3);
/*  655 */         int scale = ((Integer)record.get(4)).intValue();
/*  656 */         int ref = ((Integer)record.get(5)).intValue();
/*  657 */         int width = ((Integer)record.get(6)).intValue();
/*      */ 
/*  659 */         b.addDescriptor((short)x, (short)y, scale, ref, width, name, units);
/*      */       }
/*      */ 
/*      */     }
/*      */ 
/*  666 */     ios.close();
/*      */ 
/*  668 */     return b;
/*      */   }
/*      */ 
/*      */   private static void readBmetTableB(InputStream ios, TableB b) throws IOException {
/*      */     Document doc;
/*      */     try {
/*  674 */       SAXBuilder builder = new SAXBuilder();
/*  675 */       doc = builder.build(ios);
/*      */     } catch (JDOMException e) {
/*  677 */       throw new IOException(e.getMessage());
/*      */     }
/*      */ 
/*  680 */     Element root = doc.getRootElement();
        /* 681 */List<Element> featList = root.getChildren("featureCatalogue");
/*  682 */     for (Element featureCat : featList) {
            /* 683 */List<Element> features = featureCat.getChildren("feature");
/*      */ 
/*  685 */       for (Element feature : features) {
/*  686 */         String name = feature.getChild("annotation").getChildTextNormalize("documentation");
/*  687 */         int f = Integer.parseInt(feature.getChildText("F"));
/*  688 */         int x = Integer.parseInt(feature.getChildText("X"));
/*  689 */         int y = Integer.parseInt(feature.getChildText("Y"));
/*  690 */         int fxy = (f << 16) + (x << 8) + y;
/*      */ 
/*  692 */         Element bufrElem = feature.getChild("BUFR");
/*  693 */         String units = bufrElem.getChildTextNormalize("BUFR_units");
/*  694 */         int scale = 0; int reference = 0; int width = 0;
/*      */ 
/*  696 */         String s = null;
/*      */         try {
/*  698 */           s = bufrElem.getChildTextNormalize("BUFR_scale");
/*  699 */           scale = Integer.parseInt(clean(s));
/*      */         } catch (NumberFormatException e) {
/*  701 */           System.out.printf(" key %s name '%s' has bad scale='%s'%n", new Object[] { Integer.valueOf(fxy), name, s });
/*      */         }
/*      */         try
/*      */         {
/*  705 */           s = bufrElem.getChildTextNormalize("BUFR_reference");
/*  706 */           reference = Integer.parseInt(clean(s));
/*      */         } catch (NumberFormatException e) {
/*  708 */           System.out.printf(" key %s name '%s' has bad reference='%s' %n", new Object[] { Integer.valueOf(fxy), name, s });
/*      */         }
/*      */         try
/*      */         {
/*  712 */           s = bufrElem.getChildTextNormalize("BUFR_width");
/*  713 */           width = Integer.parseInt(clean(s));
/*      */         } catch (NumberFormatException e) {
/*  715 */           System.out.printf(" key %s name '%s' has bad width='%s' %n", new Object[] { Integer.valueOf(fxy), name, s });
/*      */         }
/*      */ 
/*  718 */         b.addDescriptor((short)x, (short)y, scale, reference, width, name, units);
/*      */       }
/*      */     }
/*      */ 
/*  722 */     ios.close();
/*      */   }
/*      */ 
/*      */   private static void readWmoXmlTableB(InputStream ios, TableB b)
/*      */     throws IOException
/*      */   {
/*      */     Document doc;
/*      */     try
/*      */     {
/*  749 */       SAXBuilder builder = new SAXBuilder();
/*  750 */       doc = builder.build(ios);
/*      */     } catch (JDOMException e) {
/*  752 */       throw new IOException(e.getMessage());
/*      */     }
/*      */ 
/*  755 */     Element root = doc.getRootElement();
        /* 756 */List<Element> featList = root
                .getChildren("BC_TableB_BUFR14_1_0_CREX_6_1_0");
/*  757 */     for (Element elem : featList)
/*      */     {
/*  759 */       String name = elem.getChildTextNormalize("ElementName_E");
/*  760 */       String units = elem.getChildTextNormalize("BUFR_Unit");
/*  761 */       int x = 0; int y = 0; int scale = 0; int reference = 0; int width = 0;
/*      */ 
/*  763 */       String fxy = null;
/*  764 */       String s = null;
/*      */       try {
/*  766 */         fxy = elem.getChildTextNormalize("FXY");
/*  767 */         int xy = Integer.parseInt(clean(fxy));
/*  768 */         x = xy / 1000;
/*  769 */         y = xy % 1000;
/*      */       }
/*      */       catch (NumberFormatException e) {
/*  772 */         System.out.printf(" key %s name '%s' has bad scale='%s'%n", new Object[] { fxy, name, s });
/*      */       }
/*      */       try
/*      */       {
/*  776 */         s = elem.getChildTextNormalize("BUFR_Scale");
/*  777 */         scale = Integer.parseInt(clean(s));
/*      */       } catch (NumberFormatException e) {
/*  779 */         System.out.printf(" key %s name '%s' has bad scale='%s'%n", new Object[] { fxy, name, s });
/*      */       }
/*      */       try
/*      */       {
/*  783 */         s = elem.getChildTextNormalize("BUFR_ReferenceValue");
/*  784 */         reference = Integer.parseInt(clean(s));
/*      */       } catch (NumberFormatException e) {
/*  786 */         System.out.printf(" key %s name '%s' has bad reference='%s' %n", new Object[] { fxy, name, s });
/*      */       }
/*      */       try
/*      */       {
/*  790 */         s = elem.getChildTextNormalize("BUFR_DataWidth_Bits");
/*  791 */         width = Integer.parseInt(clean(s));
/*      */       } catch (NumberFormatException e) {
/*  793 */         System.out.printf(" key %s name '%s' has bad width='%s' %n", new Object[] { fxy, name, s });
/*      */       }
/*      */ 
/*  796 */       b.addDescriptor((short)x, (short)y, scale, reference, width, name, units);
/*      */     }
/*  798 */     ios.close();
/*      */   }
/*      */ 
/*      */   public static TableD getWmoTableD(BufrIdentificationSection ids)
/*      */     throws IOException
/*      */   {
/*  822 */     TableD tb = (TableD)tablesD.get(version14);
/*  823 */     if (tb != null) return tb;
/*      */ 
/*  826 */     TableConfig tc14 = matchTableConfig(0, 0, 14, 0, -1);
/*  827 */     TableD result = readTableD(tc14.tableDname, tc14.tableDformat, false);
/*  828 */     tablesD.put(version14, result);
/*      */ 
/*  830 */     return result;
/*      */   }
/*      */ 
/*      */   public static TableD readTableD(String location, String format, boolean force) throws IOException {
/*  834 */     if (location == null) return null;
/*  835 */     if (location.trim().length() == 0) return null;
/*      */ 
/*  838 */     if (!force) {
/*  839 */       TableD tb = (TableD)tablesD.get(location);
/*  840 */       if (tb != null) return tb;
/*      */     }
/*      */ 
/*  843 */     InputStream ios = openStream(location);
/*  844 */     TableD d = new TableD(location, location);
/*  845 */     if (format.equals("csv")) {
/*  846 */       readWmoTableD(ios, d);
/*  847 */     } else if (format.equals("ncep")) {
/*  848 */       readNcepTableD(ios, d);
/*  849 */     } else if (format.equals("ncep-nm")) {
/*  850 */       Tables t = new Tables(null, d, null);
/*  851 */       NcepMnemonic.read(ios, t);
/*  852 */     } else if (format.equals("ecmwf")) {
/*  853 */       readEcmwfTableD(ios, d);
/*  854 */     } else if (format.equals("mel-bufr")) {
/*  855 */       readMelbufrTableD(ios, d);
/*  856 */     } else if (format.equals("wmo-xml")) {
/*  857 */       readWmoXmlTableD(ios, d);
/*      */     } else {
/*  859 */       System.out.printf("Unknown format= %s %n", new Object[] { format });
/*  860 */       return null;
/*      */     }
/*      */ 
/*  863 */     tablesD.put(location, d);
/*  864 */     return d;
/*      */   }
/*      */ 
/*      */   private static void readWmoTableD(InputStream ios, TableD tableD) throws IOException {
/*  868 */     BufferedReader dataIS = new BufferedReader(new InputStreamReader(ios, Charset.forName("UTF-8")));
/*  869 */     int count = 0;
/*  870 */     int currSeqno = -1;
/*  871 */     TableD.Descriptor currDesc = null;
/*      */     while (true)
/*      */     {
/*  874 */       String line = dataIS.readLine();
/*  875 */       if (line == null) break;
/*  876 */       if (!line.startsWith("#")) {
/*  877 */         count++;
/*      */ 
/*  879 */         if (count != 1)
/*      */         {
/*  885 */           int pos1 = line.indexOf(34);
/*  886 */           if (pos1 >= 0) {
/*  887 */             int pos2 = line.indexOf(34, pos1 + 1);
/*  888 */             StringBuffer sb = new StringBuffer(line);
/*  889 */             for (int i = pos1; i < pos2; i++)
/*  890 */               if (sb.charAt(i) == ',') sb.setCharAt(i, ' ');
/*  891 */             line = sb.toString();
/*      */           }
/*      */ 
/*  894 */           String[] flds = line.split(",");
/*  895 */           if (flds.length >= 5)
/*      */           {
/*  900 */             int fldidx = 0;
/*      */             try {
/*  902 */               int sno = Integer.parseInt(flds[(fldidx++)]);
/*  903 */               int cat = Integer.parseInt(flds[(fldidx++)]);
/*  904 */               int seq = Integer.parseInt(flds[(fldidx++)]);
/*  905 */               String seqName = flds[(fldidx++)];
/*  906 */               String featno = flds[(fldidx++)].trim();
/*  907 */               if (featno.length() != 0)
/*      */               {
/*  911 */                 String featName = flds.length > 5 ? flds[(fldidx++)] : "n/a";
/*      */ 
/*  913 */                 if (currSeqno != seq) {
/*  914 */                   int y = seq % 1000;
/*  915 */                   int w = seq / 1000;
/*  916 */                   int x = w % 100;
/*  917 */                   seqName = StringUtil.remove(seqName, 34);
/*  918 */                   currDesc = tableD.addDescriptor((short)x, (short)y, seqName, new ArrayList());
/*  919 */                   currSeqno = seq;
/*      */                 }
/*      */ 
/*  922 */                 int fno = Integer.parseInt(featno);
/*  923 */                 int y = fno % 1000;
/*  924 */                 int w = fno / 1000;
/*  925 */                 int x = w % 100;
/*  926 */                 int f = w / 100;
/*      */ 
/*  928 */                 int fxy = (f << 14) + (x << 8) + y;
/*  929 */                 currDesc.addFeature((short)fxy);
/*      */               } } catch (Exception e) {  }
/*      */ 
/*      */           }
/*      */         }
/*      */       }
/*      */     }
/*  935 */     dataIS.close();
/*      */   }
/*      */ 
/*      */   private static void readWmoXmlTableD(InputStream ios, TableD tableD)
/*      */     throws IOException
/*      */   {
/*      */     Document doc;
/*      */     try
/*      */     {
/*  953 */       SAXBuilder builder = new SAXBuilder();
/*  954 */       doc = builder.build(ios);
/*      */     } catch (JDOMException e) {
/*  956 */       throw new IOException(e.getMessage());
/*      */     }
/*      */ 
/*  959 */     int currSeqno = -1;
/*  960 */     TableD.Descriptor currDesc = null;
/*      */ 
/*  962 */     Element root = doc.getRootElement();
        /* 963 */List<Element> featList = root
                .getChildren("B_TableD_BUFR14_1_0_CREX_6_1_0");
/*  964 */     for (Element elem : featList)
/*      */     {
/*  966 */       String seqs = elem.getChildTextNormalize("FXY1");
/*  967 */       int seq = Integer.parseInt(seqs);
/*      */ 
/*  969 */       if (currSeqno != seq) {
/*  970 */         int y = seq % 1000;
/*  971 */         int w = seq / 1000;
/*  972 */         int x = w % 100;
/*  973 */         String seqName = elem.getChildTextNormalize("ElementName1_E");
/*  974 */         currDesc = tableD.addDescriptor((short)x, (short)y, seqName, new ArrayList());
/*  975 */         currSeqno = seq;
/*      */       }
/*      */ 
/*  978 */       String fnos = elem.getChildTextNormalize("FXY2");
/*  979 */       int fno = Integer.parseInt(fnos);
/*  980 */       int y = fno % 1000;
/*  981 */       int w = fno / 1000;
/*  982 */       int x = w % 100;
/*  983 */       int f = w / 100;
/*  984 */       int fxy = (f << 14) + (x << 8) + y;
/*  985 */       currDesc.addFeature((short)fxy);
/*      */     }
/*  987 */     ios.close();
/*      */   }
/*      */ 
/*      */   private static void readMelbufrTableD(InputStream ios, TableD t)
/*      */     throws IOException
/*      */   {
/*  995 */     BufferedReader dataIS = new BufferedReader(new InputStreamReader(ios));
/*  996 */     int count = 0;
/*      */     while (true)
/*      */     {
/* 1000 */       String line = dataIS.readLine();
/* 1001 */       if (line == null) break;
/* 1002 */       count++;
/*      */ 
/* 1004 */       if ((!line.startsWith("#")) && (line.length() != 0))
/*      */       {
/* 1007 */         line = line.trim();
/* 1008 */         String[] split = line.split("[ \t]+");
/* 1009 */         if (split.length >= 3) {
/* 1010 */           if (split[0].equals("END"))
/*      */             break;
/*      */           try {
/* 1013 */             short seqF = Short.parseShort(split[0]);
/* 1014 */             short seqX = Short.parseShort(split[1]);
/* 1015 */             short seqY = Short.parseShort(split[2]);
/* 1016 */             assert (seqF == 3);
/*      */ 
/* 1018 */             String seqName = "";
/* 1019 */             if (split.length > 3) {
/* 1020 */               StringBuilder sb = new StringBuilder(40);
/* 1021 */               for (int i = 3; i < split.length; i++)
/* 1022 */                 sb.append(split[i]).append(" ");
/* 1023 */               seqName = sb.toString();
/* 1024 */               seqName = StringUtil.remove(seqName, "()");
/*      */             }
/*      */ 
/* 1027 */             List seq = new ArrayList();
/*      */             while (true)
/*      */             {
/* 1030 */               line = dataIS.readLine();
/* 1031 */               if (line == null) break;
/* 1032 */               count++;
/*      */ 
/* 1034 */               if ((!line.startsWith("#")) && (line.length() != 0))
/*      */               {
/* 1037 */                 Matcher m = threeInts.matcher(line);
/*      */ 
/* 1039 */                 if (m.find()) {
/* 1040 */                   short f = Short.parseShort(m.group(1));
/* 1041 */                   short x = Short.parseShort(m.group(2));
/* 1042 */                   short y = Short.parseShort(m.group(3));
/* 1043 */                   seq.add(Short.valueOf(Descriptor.getFxy(f, x, y)));
/*      */                 }
/*      */                 else {
/* 1046 */                   m = negOne.matcher(line);
/* 1047 */                   if (m.find())
/*      */                   {
/* 1049 */                     t.addDescriptor(seqX, seqY, seqName, seq);
/* 1050 */                     break;
/*      */                   }
/*      */                 }
/*      */               }
/*      */             }
/*      */           } catch (Exception e) { log.warn(new StringBuilder().append("TableD ").append(t.getName()).append(" Failed on line ").append(count).append(" = ").append(line).append("\n ").append(e).toString()); }
/*      */         }
/*      */       }
/*      */     }
/* 1059 */     dataIS.close();
/*      */   }
/*      */ 
/*      */   private static void readNcepTableD(InputStream ios, TableD t)
/*      */     throws IOException
/*      */   {
/* 1076 */     BufferedReader dataIS = new BufferedReader(new InputStreamReader(ios));
/*      */ 
/* 1078 */     dataIS.readLine();
/*      */ 
/* 1080 */     TableD.Descriptor currDesc = null;
/*      */ 
/* 1082 */     int count = 0;
/*      */     while (true) {
/* 1084 */       String line = dataIS.readLine();
/* 1085 */       if (line == null) break;
/* 1086 */       if ((!line.startsWith("#")) && (line.trim().length() != 0))
/*      */       {
/*      */         try
/*      */         {
/* 1091 */           String[] flds = line.split("[\\|;]");
/* 1092 */           if (flds[0].equals("END"))
/*      */             break;
/* 1094 */           String fxys = flds[0].trim();
/*      */ 
/* 1096 */           if (fxys.length() > 0) {
/* 1097 */             String[] xyflds = fxys.split("-");
/* 1098 */             short x = Short.parseShort(clean(xyflds[1]));
/* 1099 */             short y = Short.parseShort(clean(xyflds[2]));
/* 1100 */             String seqName = flds.length > 3 ? flds[3].trim() : "";
/* 1101 */             currDesc = t.addDescriptor(x, y, seqName, new ArrayList());
/*      */           }
/*      */           else {
/* 1104 */             fxys = StringUtil.remove(flds[1], 62);
/* 1105 */             String[] xyflds = fxys.split("-");
/* 1106 */             short f = Short.parseShort(clean(xyflds[0]));
/* 1107 */             short x = Short.parseShort(clean(xyflds[1]));
/* 1108 */             short y = Short.parseShort(clean(xyflds[2]));
/* 1109 */             int fxy = (f << 14) + (x << 8) + y;
/* 1110 */             currDesc.addFeature((short)fxy);
/*      */           }
/*      */ 
/*      */         }
/*      */         catch (Exception e)
/*      */         {
/* 1118 */           log.error(new StringBuilder().append("Bad table ").append(t.getName()).append(" entry=<").append(line).append(">").toString(), e);
/*      */         }
/*      */       }
/*      */     }
/*      */ 
/* 1123 */     dataIS.close();
/*      */   }
/*      */ 
/*      */   private static void readEcmwfTableD(InputStream ios, TableD t)
/*      */     throws IOException
/*      */   {
/* 1135 */     BufferedReader dataIS = new BufferedReader(new InputStreamReader(ios));
/*      */ 
/* 1137 */     TableD.Descriptor currDesc = null;
/*      */ 
/* 1139 */     int n = 0;
/*      */     while (true) {
/* 1141 */       String line = dataIS.readLine();
/* 1142 */       if (line == null) break;
/* 1143 */       line = line.trim();
/* 1144 */       if ((!line.startsWith("#")) && (line.length() != 0))
/*      */       {
/*      */         try
/*      */         {
/* 1150 */           String[] flds = line.split("[\\s]+");
/*      */           String fxys;
/* 1151 */           if (n == 0) {
                        /* 1152 */fxys = flds[0].trim();
/* 1153 */             int fxy = Integer.parseInt(fxys);
/* 1154 */             int y = fxy % 1000;
/* 1155 */             fxy /= 1000;
/* 1156 */             int x = fxy % 100;
/* 1157 */             currDesc = t.addDescriptor((short)x, (short)y, "", new ArrayList());
/*      */ 
/* 1159 */             n = Integer.parseInt(flds[1]);
/* 1160 */             fxys = flds[2].trim();
/*      */           } else {
/* 1162 */             fxys = flds[0].trim();
/*      */           }
/*      */ 
/* 1165 */           int fxy = Integer.parseInt(fxys);
/* 1166 */           int y = fxy % 1000;
/* 1167 */           fxy /= 1000;
/* 1168 */           int x = fxy % 100;
/* 1169 */           int f = fxy /= 100;
/* 1170 */           fxy = (f << 14) + (x << 8) + y;
/* 1171 */           currDesc.addFeature((short)fxy);
/* 1172 */           n--;
/*      */         }
/*      */         catch (Exception e)
/*      */         {
/* 1178 */           log.error(new StringBuilder().append("Bad table ").append(t.getName()).append(" entry=<").append(line).append(">").toString(), e);
/*      */         }
/*      */       }
/*      */     }
/* 1181 */     dataIS.close();
/*      */   }
/*      */ 
/*      */   private static InputStream open(String location)
/*      */     throws IOException
/*      */   {
/* 1187 */     InputStream ios = null;
/*      */ 
/* 1190 */     String tmp = new StringBuilder().append("/resources/bufrTables/local/").append(location).toString();
/* 1191 */     ios = BufrTables.class.getResourceAsStream(tmp);
/* 1192 */     if (ios != null)
/*      */     {
/* 1194 */       return ios;
/*      */     }
/*      */ 
/* 1197 */     if (location.startsWith("http:")) {
/* 1198 */       URL url = new URL(location);
/* 1199 */       ios = url.openStream();
/*      */     } else {
/* 1201 */       ios = new FileInputStream(location);
/*      */     }
/* 1203 */     return ios;
/*      */   }
/*      */ 
/*      */   static InputStream openStream(String location) throws IOException {
/* 1207 */     InputStream ios = null;
/*      */ 
/* 1209 */     if (location.startsWith("resource:")) {
/* 1210 */       location = location.substring(9);
/* 1211 */       ios = BufrTables.class.getResourceAsStream(location);
/* 1212 */       if (ios == null)
/* 1213 */         throw new RuntimeException(new StringBuilder().append("resource not found=<").append(location).append(">").toString());
/* 1214 */       return ios;
/*      */     }
/*      */ 
/* 1217 */     if (location.startsWith("http:")) {
/* 1218 */       URL url = new URL(location);
/* 1219 */       ios = url.openStream();
/*      */     } else {
/* 1221 */       ios = new FileInputStream(location);
/*      */     }
/* 1223 */     return ios;
/*      */   }
/*      */ 
/*      */   public static void main(String[] args)
/*      */     throws IOException
/*      */   {
/* 1229 */     Formatter out = new Formatter(System.out);
/*      */ 
/* 1231 */     TableB tableB = getWmoTableB(13);
/* 1232 */     tableB.show(out);
/*      */ 
/* 1234 */     TableD tableD = getWmoTableD(null);
/* 1235 */     tableD.show(out);
/*      */   }
/*      */ 
/*      */   public static class Tables
/*      */   {
/*      */     public TableB b;
/*      */     public TableD d;
/*      */     public BufrTables.Mode mode;
/*      */ 
/*      */     Tables()
/*      */     {
/*      */     }
/*      */ 
/*      */     Tables(TableB b, TableD d, BufrTables.Mode mode)
/*      */     {
/*  361 */       this.b = b;
/*  362 */       this.d = d;
/*  363 */       this.mode = (mode == null ? BufrTables.Mode.wmoOnly : mode);
/*      */     }
/*      */   }
/*      */ 
/*      */   private static class TableConfig
/*      */   {
/*      */     int center;
/*      */     int subcenter;
/*      */     int master;
/*      */     int local;
/*      */     int cat;
/*      */     String tableBname;
/*      */     String tableBformat;
/*      */     String tableDname;
/*      */     String tableDformat;
/*  211 */     BufrTables.Mode mode = BufrTables.Mode.wmoLocal;
/*      */ 
/*      */     boolean matches(int center, int subcenter, int master, int local, int cat) {
/*  214 */       if ((this.center >= 0) && (center >= 0) && (center != this.center)) return false;
/*  215 */       if ((this.subcenter >= 0) && (subcenter >= 0) && (subcenter != this.subcenter)) return false;
/*  216 */       if ((this.master >= 0) && (master >= 0) && (master != this.master)) return false;
/*  217 */       if ((this.local >= 0) && (local >= 0) && (local != this.local)) return false;
/*  218 */       if ((this.cat >= 0) && (cat >= 0) && (cat != this.cat)) return false;
/*  219 */       return true;
/*      */     }
/*      */     public String toString() {
/*  222 */       return this.tableBname;
/*      */     }
/*      */   }
/*      */ 
/*      */   public static enum Mode
/*      */   {
/*  120 */     wmoOnly, 
/*  121 */     wmoLocal, 
/*  122 */     localOverride;
/*      */   }
/*      */ }

/* Location:           /home/bclement/workspace/AWIPS2_baseline/cots/ucar.nc2/bufr-3.0.jar
 * Qualified Name:     ucar.nc2.iosp.bufr.tables.BufrTables
 * JD-Core Version:    0.6.2
 */